"""
Django-partial-page is a middleware that allows your views send the whole pages as usual or just parts of the same pages.

Add the middleware to settings.MIDDLEWARE_CLASSES. For it to work,
1. The process_views of the middleware above should not short-circuit by calling the view prematurely.
2. The middleware below should not substitute the TemplateResponse with something else.

If other middleware changes directly response.content, these changes will be ignored.
"""



from functools import wraps
from threading import local

from django.conf import settings
from django.http import HttpResponse
from django.utils import simplejson as json


PARTIAL_PARAM_NAME = getattr(settings, 'PARTIAL_PARAM_NAME', 'partial')
_thread_locals = local()


def capture_blocks(func):
	"""
	Decorates .render() methods of nodes to capture their output and save it into _thread_locals.partials['blocks']

	The node class must have a `name` attribute.
	"""
	@wraps(func)
	def decorated(self, context):
		result = func(self, context)
		if not _thread_locals.request.no_partial and self.name in getattr(_thread_locals.request, 'partials', []):  # save the block contents if it's requested in GET
			_thread_locals.partial['blocks'][self.name] = result  # this overwrites the existing value, but that's ok, since the final block is what is shown in the rendered template.
		return result

	return decorated


class PartialPageMiddleware(object):
	"""
	Catches the requests for partial pages and saves all the named blocks in a dictionary, then sends the dictionary as JSON.

	Notes: this middleware will catch the blocks contents at the moment they are rendered, no matter who calls response.render(). But if other middleware classes replace the response contents (say, redirect links to a counter), these changes can't be saved in the dictionary.
	"""
	def __init__(self):
		"""
		Monkey patches BlockNode so that if block name is requested in GET parameters,
		it will be saved in _thread_locals and sent in JSON response.
		"""
		from django.template.loader_tags import BlockNode

		# here we stick our nose into block node:
		# we capture its return value and save it into a dictionary
		BlockNode.render = capture_blocks(BlockNode.render)

	def process_view(self, request, callback, callback_args, callback_kwargs):
		"""
		Tests if partial parameter is in GET parameters and extracts block names from it.
		"""
		_thread_locals.request = request
		request.no_partial = callback_kwargs.pop('no_partial', False)
		request.partials = []
		if not request.no_partial and PARTIAL_PARAM_NAME in request.GET:
			request.partials = request.GET[PARTIAL_PARAM_NAME].split(',')

		_thread_locals.partial = {'blocks': {}, 'data': {}}

	def process_response(self, request, response):
		"""
		Bypasses the response or, if partial page is requested, returns JSON.
		"""
		if getattr(request, 'no_partial', False) or len(getattr(request, 'partials', [])) == 0:
			return response

		json_obj = dict(
			(key, _thread_locals.partial[key])
			for key in ['blocks', 'data']
			if len(_thread_locals.partial[key]) > 0)

		new_response = HttpResponse(json.dumps(json_obj), mimetype='application/json')
		new_response.status_code = response.status_code
		return new_response


def set_js_data(**kwargs):
	"""
	Updates the `data` dictionary in the json object that partial middleware sends.
	"""
	_thread_locals.partial['data'].update(kwargs)


def no_partial(func):
	"""
	A decorator that protects a view from partial page requests.
	(Unless the view itself overrides this.)
	"""
	@wraps(func)
	def _decorated(request, *args, **kwargs):
		request.no_partial = True
		return func(request, *args, **kwargs)
	return _decorated



from setuptools import setup, find_packages

setup(
    name='django-partial-page',
    version='0.4.1',
    url='https://bitbucket.org/siberiano/django-partial-page/',
    license='Public Domain',
    author='Dmitri Lebedev',
    author_email='detail@ngs.ru',
    description="""Middleware that extracts {% block-s %} from pages and sends them as JSON. This allows the clients update parts of the pages, which is useful for Ajax apps and sites using History.PushState.
    
    If /mypage/ has {% block main_content %} in the template, make a request /mypage/?partial=main_content, and you'll receive a JSON: {"main_content": "..."}.
    
    The example Django project in the BitBucket contains a working JavaScript module that handles these requests and does something more.
    """,
    include_package_data=True,
    zip_safe=False,
    package_dir={'': 'example'},
    packages=find_packages('example', exclude=('*.pyc', '*~'))
)



from django.template import Library, TemplateSyntaxError
from django.template.loader_tags import BlockNode
from django.utils import simplejson as json
from django.utils.safestring import mark_safe
from django_partial_page.middleware import _thread_locals, capture_blocks

register = Library()


@register.simple_tag
def partial_js_data():
	if len(_thread_locals.partial['data']) == 0:
		return ''

	return mark_safe('<script>require(["partial"], function (p) {{ p.data = {0}; }})</script>'.format(json.dumps(_thread_locals.partial['data'])))


class DelayedBlockNode(BlockNode):
	"""
	DelayedBlockNode can have a stub part:
	{% delayed_block block1 %}
		real content that's too big to send immediately
	{% delayed_stub %}
		stub content that's short and can be shown right away
	{% enddelayed_block %}

	If the request is normal, delayed block will render the stub content, if present. If the request is partial, it will render the normal content.

	You can mix delayed blocks with those normal, the end result will be what is rendered in the lowermost level. The {{ block.super }} context variable will be different depending on what request is sent.
	"""
	def __init__(self, name, nodelist, nodelist_stub):
		self.nodelist_real = nodelist
		self.nodelist_stub = nodelist_stub
		super(DelayedBlockNode, self).__init__(name, nodelist)

	@capture_blocks
	def render(self, context):
		if self.name in getattr(_thread_locals.request, 'partials', []):
			self.nodelist = self.nodelist_real
			return super(DelayedBlockNode, self).render(context)
		
		self.nodelist = self.nodelist_stub
		return mark_safe('<div id="partial_{0}" class="partial_delayed">{1}</div>'.format(self.name, super(DelayedBlockNode, self).render(context)))


@register.tag
def delayed_block(parser, token):
	"""
	Define a block that can be overridden by child templates and delayed (rendered later).

	This code is a mix of django.template.loader_tags.do_block and django.template.default_tags.do_for.
	"""
	bits = token.contents.split()
	if len(bits) != 2:
		raise TemplateSyntaxError("'%s' tag takes only one argument" % bits[0])
	block_name = bits[1]

	# Keep track of the names of BlockNodes/DelayedBlockNodes found in this template, so we can
	# check for duplication.
	try:
		if block_name in parser.__loaded_blocks:
			raise TemplateSyntaxError("'%s' tag with name '%s' appears more than once" % (bits[0], block_name))
		parser.__loaded_blocks.append(block_name)
	except AttributeError: # parser.__loaded_blocks isn't a list yet
		parser.__loaded_blocks = [block_name]

	nodelist = parser.parse(('delayed_stub', 'enddelayed_block', 'enddelayed_block %s' % block_name))
	token = parser.next_token()

	if token.contents == 'delayed_stub':
		nodelist_stub = parser.parse(('enddelayed_block',))
		parser.delete_first_token()
	else:
		nodelist_stub = None

	parser.delete_first_token()
	return DelayedBlockNode(block_name, nodelist, nodelist_stub)
